require "primitive_wrapper/version"
require "blockify"
require "yieldhelper"  # am I using this really?
require "pred"

#
#  Standard Object upgrades
#
class Object
  def type_of?(cls)
    kind_of? cls
  end
  def type
    self.class
  end
  def pw_copy
    self.dup rescue self
  end
  def prim_value
    self
  end
  def wrapped?
    false
  end
  def to_wrapper
    return Bit.new(nil) if self.nil?
    return Bool.new(true) if self==true
    return Bool.new(false) if self==false
    return Int.new(self) if self.kind_of? Integer
    return FloatW.new(self) if self.kind_of? Float
    return Number.new(self) if self.kind_of? Numeric
    return Datum.new(self) if self.kind_of? String
    return SymbolW.new(self) if self.kind_of? Symbol
    return Property.new(self) if self.kind_of? Hash
    return XArray.new(self) if self.kind_of? Array
    return XRange.new(self) if self.kind_of? Range
    return Fraction.new(self) if self.kind_of? Rational
    return Value.new(self)
  end
end

class Symbol
  def ~
    self
  end
end

class TrueClass
  def ~
    self
  end
end

class FalseClass
  def ~
    self
  end
end

class NilClass
  def ~
    self
  end
end

class Range
  def ~
    self
  end
end

class Array
  def ~
    self
  end
end

class Range       # if it makes sense, add methods here 
  def reversed? 
    first > last
  end
  def to_xr
    XRange.new self
  end
  def reverse
    if exclude_end?
      if (first==last)
        (last...first)
      else
        ((last.pred)..first)
      end
    else
      (last..first)
    end    
  end
  def reorder
    if exclude_end?
      last < first ? reverse : (first...last)
    else
      last < first ? reverse : (first..last)
    end
  end
  def reversed?
    last < first 
  end
  def simplify
    if exclude_end?
      return (first...last) if first==last
      return (first..last.pred)
    end
    return (first..last)
  end
  def element_class
    return first.class
  end
  def simple?
    return true if first.type_of? Integer
    if first.type_of? String
      return true if first.length==1
    end
    false
  end
end

class Array
  def empty!
    old = self.dup
    replace []
    return old
  end
  def to_xa
    XArray.new(self)
  end
end

#
#  BASE CLASS == Value ... a wrapper for any ruby object
#
class Value
  def self.raise_on_freeze
   @freeze_raise = true
  end
  def self.ignore_on_freeze
   @freeze_raise = false
  end
  def self.freeze_raise?
    @freeze_raise
  end
  ignore_on_freeze
  
  def initialize(obj=nil)
    obj = obj.prim_value
    ensure_valid(obj)
    @value = obj
  end
  
  def to_wrapper
    self
  end

  def wrapped?
    true
  end

  def type_of?(cls)
    return true if @value.kind_of? cls
    return true if self.kind_of? cls
    false
  end
  
  def type
    @value.class
  end  
  
  def ensure_valid(obj, mess = "Incompatible type")
    unless valid_type(obj)
      raise mess
    end  
  end 

  def freeze  # never allow the wrapper to be frozen
    raise "this object cannot be frozen" if self.class.freeze_raise?
  end

  def val
    @value
  end
  
  def ~
    @value
  end
  
  def val=(dat)
    replace(dat)
  end

  def prim_value
    @value
  end
  
  def unwrap
    @value
  end

  def ==(other)
    other.prim_value == @value
  end

  def !=(other)
    other.prim_value != @value
  end  

  def valid_type(prm=nil)
    true
  end
  
  def replace(other)    
    @value = other.prim_value if valid_type(other)
  end
  
  def to_s
    @value.to_s
  end  

  def inspect
    str = @value.nil? ? "nil" : @value.inspect
    "(#{self.class}==>#{str})"
  end  
end

class Bool < Value
  def valid_type(prm)
    return true if prm.kind_of? TrueClass
    return true if prm.kind_of? FalseClass
    return true if prm.kind_of? TrueW
    return true if prm.kind_of? FalseW
    return true if prm.kind_of? Bool
    false
  end
  
  def |(other)
    Bool.new(@value | other.prim_value)
  end
  
  def &(other)
    Bool.new(@value & other.prim_value)
  end
  
  def ^(other)
    Bool.new(@value ^ other.prim_value)
  end

  def !
    Bool.new(!@value)
  end
    
  def to_i
    @value ? 1:0
  end
  def to_int
    Int.new(@value ? 1:0)
  end
end

class Bit < Bool
  def valid_type(prm)
    return true if prm.kind_of? TrueClass
    return true if prm.kind_of? FalseClass
    return true if prm.kind_of? NilClass
    return true if prm.kind_of? Bool
    false
  end 
  def |(other)
    @value | other.prim_value
  end
  
  def &(other)
    @value & other.prim_value
  end
  
  def ^(other)
    @value ^ other.prim_value
  end

  def !
    !@value
  end
  def to_i
    @value ? 1:0
  end
  def to_int
    Int.new(@value ? 1:0)
  end  
end

class FixedBit < Bit
  def initialize
    raise "FixedBit cannot create instance"
  end    
  def replace(other)
    raise "can't assign primitive type"
  end  
end

class FalseW < FixedBit
  def initialize
    @value = false
  end    
end

class TrueW < FixedBit
  def initialize
    @value = true  
  end
end

class Null < FixedBit
  def initialize
    @value = nil
  end
end

class SymbolW < Value
  def valid_type(prm)
    return true if prm.kind_of? Symbol
    return true if prm.kind_of? SymbolW
    false
  end
  
  FIX_ARGS = lambda do |elm|
    rtn = elm
    if elm.kind_of? Symbol
      rtn = elm.to_s
    elsif elm.kind_of? SymbolW
      rtn = elm.to_s
    end
    rtn
  end
    
  def self.bestow_string_mutate_method(meth_def_name, meth_call_name)
    define_method meth_def_name do |*args, &block| 
      prms = args.blockify_elements &FIX_ARGS
      str = @value.to_s
      str.send(meth_call_name, *prms, &block)
      @value = str.to_sym
    end  
  end
  
  def self.bestow_string_non_mutate_method(meth_def_name, meth_call_name, return_type = SymbolW)
    if return_type==Symbol
      define_method meth_def_name do |*args, &block|
        prms = args.blockify_elements &FIX_ARGS
        @value.to_s.send(meth_call_name, *prms, &block).to_sym
      end
    elsif return_type==String
      define_method meth_def_name do |*args, &block|
        prms = args.blockify_elements &FIX_ARGS
        @value.to_s.send(meth_call_name, *prms, &block).to_s
      end
    elsif return_type==SymbolW
      define_method meth_def_name do |*args, &block|
        prms = args.blockify_elements &FIX_ARGS
        SymbolW.new @value.to_s.send(meth_call_name, *prms, &block).to_s.to_sym
      end
    else  # default type
      define_method meth_def_name do |*args, &block|
        prms = args.blockify_elements &FIX_ARGS
        @value.to_s.send(meth_call_name, *prms, &block)
      end
    end    
  end 
  
  def self.bestow_symbol_method(meth_def_name, meth_call_name, mutate_self=false)
    if mutate_self      
      define_method meth_def_name do |*args, &block| 
        margs = args.blockify_elements { |t| t.prim_value }       
        tval = @value.send(meth_call_name, *margs, &block)
        ensure_valid(tval, "symbol mutate method not valid")
        @value = tval
        return self
      end  
    else
      define_method meth_def_name do |*args, &block| 
        margs = args.blockify_elements { |t| t.prim_value }
        rtn = @value.send(meth_call_name, *margs, &block)
        return rtn
      end  
    end
  end 
  
  def self.bestow_string_mutate_methods(meths)
    meths.each do |meth|
      bestow_string_mutate_method(meth, meth)
    end    
  end
  
  def self.bestow_string_non_mutate_methods(meths, return_type = SymbolW)
    meths.each do |meth|
      bestow_string_non_mutate_method(meth, meth, return_type)
    end    
  end
    
  def self.bestow_symbol_methods(meths, mutate_self=false)
    meths.each do |meth|
      dmeth=meth 
      if mutate_self
        dmeth = (meth.to_s + '!').to_sym
      end
      bestow_symbol_method(dmeth, meth, mutate_self)
    end    
  end
  
  # create non-mutate SymbolW methods
  bestow_symbol_methods( [:next, :succ, :pred, :[], :length, :size, :upcase, :downcase, :capitalize, :swapcase, :<, :>, :<=, :>=, :between?, :empty? ], false)
  
  # create mutate SymbolW methods  ... these will auto-bang!
  bestow_symbol_methods([:next, :succ, :upcase, :downcase, :capitalize, :swapcase, :pred], true)
  
  # string non-mutate methods
  bestow_string_non_mutate_methods([:match, :include?], nil)
  bestow_string_non_mutate_methods([:prepend], SymbolW)
  bestow_string_non_mutate_method(:append, :<<, SymbolW)

  # string mutate methods
  bestow_string_mutate_method(:prepend!, :prepend)
  bestow_string_mutate_method(:append!, :<<)
  bestow_string_mutate_method(:[]=, :[]=)

  # why did I redefine this? don't remember ... investigate later
  def include? pat
    self.to_s.include? pat.to_s
  end
  
  # old-fashioned defines:
  def to_sym
    @value
  end
  def to_str
    @value.to_s
  end
end

class ValueAdd < Value
  def valid_type(prm) # must override
    false
  end
  def self.bestow_methods(*args)
    args = args.first if args.first.kind_of? Array
    args.each do |meth|
      define_method meth do |*args, &block|
        @value.send(meth, *args, &block)
      end  
    end    
  end  
  def self.capture_base_methods(type, except=Object)
    add_me = type.instance_methods - except.instance_methods - Value.instance_methods - [:singleton_method_added]
    bestow_methods add_me
  end
end

class Int < ValueAdd
  capture_base_methods(5.class)  # Bignum methods same as Fixnum ... updated ruby no longer uses Fixnum
  def valid_type(prm)
    return true if prm.kind_of? Integer
    return true if prm.kind_of? Int
    false
  end
  def inc   # simulates post-increment
    @value+= 1
    return @value-1
  end
  def dec   # simulates post-deccrement
    @value-= 1
    return @value+1
  end
  def to_int
    self
  end
  def ~
    ~@value
  end
  
  def pre_inc
    @value+=1 
  end

  def pre_dec  
    @value-=1 
  end
end

class FloatW < ValueAdd
  capture_base_methods(Float)
  def valid_type(prm)
    return true if prm.kind_of? Float
    return true if prm.kind_of? FloatW
    false
  end
  def to_int
    Int.new @value.to_i
  end
end

class Number < ValueAdd
  capture_base_methods(5.class)
  capture_base_methods(Float, self)
  capture_base_methods(Complex, self)
  capture_base_methods(Rational, self)  
  def valid_type(prm)
    return true if prm.kind_of? Numeric
    return true if prm.kind_of? Number
    false
  end
  def ~   # make sure capture did not muck things up
    @value   # If you need ~ use Int instead of Number
  end
  def to_int
    Int.new @value.to_i
  end  
end

# string, number
class Datum < ValueAdd
  capture_base_methods(Number)
  capture_base_methods(String, self)
  def valid_type(prm)
    return true if prm.kind_of? Numeric
    return true if prm.kind_of? String
    return true if prm.kind_of? Datum
    false
  end

  # methods that broke ... fixed here  
  def to_s
    @value.to_s    
  end
  def ~   # make sure capture did not muck things up
    @value   # If you need ~ use Int instead of Number
  end 
  def to_int
    Int.new @value.to_i
  end  
end

class Property < Value
  GOOD_KEY_RGX = /\A[a-zA-Z_][a-zA-Z_0-9]*\z/

  def initialize(hash={})
    self.val=(hash)
  end
  
  def self.bad_key?(idx)
    @badkeys ||= {}
    true==@badkeys[idx]
  end

  def self.good_candidate_name? name
    name = name.prim_value
    return false unless name.kind_of? Symbol
    tst = GOOD_KEY_RGX.match name
    return false if tst.nil?
    true
  end
  
  def self.good_key?(idx)  # syntax and not bad ... does not check to see if taken as only instansts can do that
    idx = idx.prim_value
    return false unless good_candidate_name? idx 
    return false if bad_key?(idx)
    true    
  end
  
  def self.reserve_property_names!(*names)
    @badkeys ||= {}
    names = names.first if names.first.kind_of? Array
    names.each do |key|
      next if key[-1]=='!'
      next if key[-1]=='?'
      kw = key[-1]=='=' ? key[0..-2] : key[0..-1]
      md = GOOD_KEY_RGX.match kw
      next if md.nil?  # bad key 
      @badkeys[key]=true
    end
  end
  reserve_property_names!(self.instance_methods)
  reserve_property_names!(Object.new.private_methods)
  
  # assign replacement hash
  def val=(hash)
    ensure_valid(hash)
    hash = hash.prim_value
    @value = hash    # keep this as prime as possible
    rekey!
  end
  
  def rekey!
    @good_keys = {}  # these keys can be properties
    @hold_keys  = {} # these keys can't
    @value.each_pair do |pkey,val|
      key = pkey.prim_value
      if self.class.good_key? key
        @good_keys[key.prim_value]=true
      else
        @hold_keys[key.prim_value]=true
      end
    end
  end  
  
  def method_missing(method, *args, &block)
    flag_good = @good_keys.include? method
    unless flag_good
      unless method[-1]=='='
        if @value.include? method
          rekey!
          flag_good = @good_keys.include? method
        end
      end
    end
    raise "Illegal Block attachment" unless block.nil?
    if flag_good
      raise "*** ArgumentError Exception: wrong number of arguments (given #{args.count}, expected 0)" unless args.empty?
      return @value[method]  
    end
    if method[-1]=='='  # assignment
      key = method[0..-2].to_sym      
      if args.count==1
        if @good_keys.include? key
          @value[key] = args[0]
          return true
        end
        # creating a new key here ... good or bad?
        if self.class.good_key? key
          @good_keys[key]=true
        else
          @hold_keys[key]=true
        end
        return @value[key] = args[0] 
      end
      raise "*** ArgumentError Exception: wrong number of arguments (given #{args.count}, expected 1)"
    else
      # lookup failed ... punt
      super
    end
  end
    
  def split!
    good = {}
    bad = {}
    @good_keys.each_pair do |key,val|    
      good[key] = @value[key]
    end
    @hold_keys.each_pair do |key,val|
      bad[key] = @value[key]
    end
    return [good, bad]
  end
  
  def [](idx)
    @value[idx]
  end
  
  def []=(idx,val)
    idx = idx.prim_value # in case it is wrapped
    rtn = @value[idx] = val
    # now put key in one of two bins
    if self.class.good_key? idx
      @good_keys[idx]=true
    else
      @hold_keys[idx]=true
    end
    return rtn
  end

  def import_hash!(hash)
    hash.each_pair do |key,val|
      self[key]=val
    end 
  end
  
  def valid_type(prm)
    return true if prm.kind_of? Hash
    return true if prm.kind_of? Property
    false
  end
  
  def defined_properties!
    rekey!
    @good_keys.keys.sort
  end
  def deferred_properties!
    rekey!
    @hold_keys.keys  # can't sort bad keys without a ton of work
  end
  
  def property?(tst)
    rekey!
    tst = tst.prim_value
    return true if @good_keys.include? tst
    false
  end

  def deferred?(tst)
    rekey!
    tst = tst.prim_value
    return true if @hold_keys.include? tst
    false
  end
  
  def define_properties!(ary, dflt="")
    ary.each do |item|
      key = item.prim_value
      raise "*** Property Error: must use Symbol type to define properties" unless key.kind_of? Symbol
      raise "*** Property Error: property #{key.inspect} already defined" if @good_keys.include? key
      raise "*** Property Error: property #{key.inspect} violates naming convention" unless self.class.good_candidate_name? key
      raise "*** Property Error: property #{key.inspect} is reserved" if self.class.bad_key? key
      @good_keys[key]=true
      @value[key]=dflt
    end
  end
end

class Fraction < ValueAdd
  capture_base_methods(Rational)
  
  def coerce(other)
    [Fraction.new(other.prim_value), self.prim_value]
  end  
  
  def self.approximate_by_float(p1, pcnt=nil)
    return Fraction.new(p1.prim_value) if pcnt.nil?
    pcnt/= 100.0
    neg = p1.negative?
    p1=p1.abs
    if p1 >=1.0
      top = p1.floor
      bot = 1
    else
      top = 1
      bot = (1/p1).floor
    end
    loop do
      app = top.to_f/bot.to_f
      tst = (app - p1).abs/p1
      break if tst < pcnt
      if app > p1
        bot += 1
      else
        top += 1
      end        
    end
    frac = Fraction.new(top.prim_value,bot.prim_value)
    frac = -frac if neg
    return frac
  end

  def initialize(*prms)
    #obj = obj.prim_value
    #ensure_valid(obj)
    #@value = obj
    if 1==prms.count
      @value = Rational(prms[0].prim_value)
    elsif 2==prms.count
      if prms[0].type_of? Float
        if prms[1].type_of? Integer
          sign = prms[0].negative? ? -1:1
          num = prms[0].abs
          whole = num.floor
          frac = num-whole
          num = (frac * prms[1]).round
          frac = Fraction.new(whole, num, prms[1]) * sign
          @value = frac.prim_value
        else
          fract = self.class.approximate_by_float(prms[0].prim_value,prms[1].prim_value)
          @value = fract.prim_value
        end
      else
        @value = Rational(prms[0].prim_value,prms[1].prim_value)
      end
      
    elsif 3==prms.count
      @value = Rational(prms[1].prim_value,prms[2].prim_value) + prms[0].prim_value
    else
      raise("expected 1..3 parameters")
    end
  end
  
  def replace(other)    
    @value = Rational(other.prim_value)
  end
  
  def abs!
    @value = @value.abs
    self
  end

  def negate!
    @value = -@value.abs
    self
  end
  
  def +(other)
    Fraction.new(@value + other)
  end

  def -(other)
    Fraction.new(@value - other)
  end

  def *(other)
    Fraction.new(@value * other)
  end
  
  def **(other)
    Fraction.new(@value ** other)
  end

  def %(other)
    Fraction.new(@value % other)
  end

  def /(other)
    Fraction.new(@value / other)
  end

  def +@
    Fraction.new(@value)
  end

  def -@
    Fraction.new(-@value)
  end
  
  def to_fraction
    self
  end

  def to_s(paren=true)
    return @value.numerator if @value.denominator==1
    if @value.numerator.abs < @value.denominator
      if paren
        return "(#{@value.numerator}/#{@value.denominator})"
      else
        return "#{@value.numerator}/#{@value.denominator}"
      end    
    end    
    if @value.negative?
      rat = -@value
      whole, rem = rat.numerator.divmod rat.denominator
      if paren
        return "(-#{whole} -#{rem}/#{@value.denominator})"
      else
        return "-#{whole} -#{rem}/#{@value.denominator}"
      end
      
    end
    whole, rem = @value.numerator.divmod @value.denominator
    if paren
      return "(#{whole} #{rem}/#{@value.denominator})"
    else
      return "#{whole} #{rem}/#{@value.denominator}"
    end
  end
  
  def valid_type(prm)
    return true
  end
  
end

module Kernel
  def Fraction(*prms)
    Fraction.new(*prms)
  end
end

class Integer
  def to_fraction
    Fraction.new(self)
  end
end

class Float
  def to_fraction(pcnt=nil)
    Fraction.approximate_by_float(self, pcnt)
  end
end

class Rational
  def to_fraction
    Fraction.new(self)
  end
end

class XRange < ValueAdd
  capture_base_methods(Range)
  def valid_type(prm)
    return true if prm.kind_of? Range
    return true if prm.kind_of? XRange
    false
  end
  def reverse!
    @value = @value.reverse
    self
  end
  def reverse
    @value.reverse.to_xr
  end
  def reorder!
    @value = @value.reorder
    self
  end
  def simplify!
    @value = @value.simplify
    self
  end  
  def simplify
    @value.simplify.to_xr
  end  
  def to_xr
    self
  end
  def to_range  # add similar to other wrappers
    @value
  end
  def re_range(ng) 
    return self if ng.nil?
    return self if ng <= 0
    return self unless first.kind_of? Integer    
    t_first = first
    t_last = last
    if first.negative?
      t_first = ng+first
    end
    if last.negative?
      t_last = ng+last
    end
    if exclude_end?
      XRange.new (t_first...t_last)
    else
      XRange.new (t_first..t_last)
    end
  end
  def re_range!(ng)
    replace re_range(ng)
  end  
  def size   # getting the wrong count ... off by 1
    dirty = false
    sz = @value.size
    if sz.nil?
      dirty = true
    elsif sz==0
      dirty = true
    end
    unless dirty
      return sz
    end
    if first.respond_to? :lcm  # Covers both Integer and Int
      sz = reversed? ? @value.reverse.size : @value.size
      return sz
    end
    range = @value.reorder
    cnt = 0
    # curr = range.first.dup rescue range.first
    curr = range.first.pw_copy
    stop = range.last
    loop do
      cnt += 1  # 5..5 is count of 1
      break if curr==stop
      curr.succ!
    end
    cnt -= 1 if exclude_end?
    return cnt  
  end
  def count
    size
  end
  def include?(val)  # account for reversed ordering
    @value.reorder.include? val
  end
  def cover?(val)  # means val>=start and val<=end ... different than include
    @value.reorder.include? val
  end
  def member?(val) # same as include?
    @value.reorder.include? val
  end
  def to_a
    dat = @value.reorder.to_a
    return reversed? ? dat.reverse : dat    
  end
  def max(*n, &block)
    @value.reorder.send(:max, *n, &block)  
  end
  def min(*n, &block)
    @value.reorder.send(:min, *n, &block)  
  end
  def eql?
    @value.eql? (~other)
  end  
  def ===(other)
    @value.include? other
  end
  
  def ==(other)
    @value.eql? (~other)
  end
    
  def step(n=1)
    n = 1 if n<= 0
    return Enumerator.enumerate_yields(self, :step, n) unless block_given?
    curr = first.pw_copy
    if exclude_end?
      if reversed?
        loop do
          break if curr <= last
          yield curr
          n.times { |t| curr = curr.pred }
        end
      else
        loop do
          break if curr >= last
          yield curr
          n.times { |t| curr = curr.succ }
        end
      end
    else
      if reversed?
        loop do
          yield curr
          break if curr <= last
          n.times { |t| curr = curr.pred }
        end
      else
        loop do
          yield curr
          break if curr >= last
          n.times { |t| curr = curr.succ }
        end
      end
    end    
  end

  def reverse_step(n=1)
    n = 1 if n<= 0
    return Enumerator.enumerate_yields(self, :reverse_step, n) unless block_given?
    curr = last.pw_copy
    if exclude_end?
      if reversed?
        loop do
          break if curr >= first
          yield curr
          n.times { |t| curr = curr.succ }
        end
      else
        loop do
          break if curr <= first
          yield curr
          n.times { |t| curr = curr.pred }
        end
      end
    else
      if reversed?
        loop do
          yield curr
          break if curr >= first
          n.times { |t| curr = curr.succ }
        end
      else
        loop do
          yield curr
          break if curr <= first
          n.times { |t| curr = curr.pred }
        end
      end
    end    
  end

  def each(&block)
    self.send(:step, 1, &block)
  end 
  
  def reverse_each(&block)
    self.send(:reverse_step, 1, &block)
  end 

end

module PrimitiveWrapper
  def self.get_list(list, ng=nil)
    rtn = []
    if list.empty?
      return []
    elsif list.count == 1
      item = list.first
      if item.type_of? Range
        rtn.push item.to_wrapper
      else
        rtn.push item
      end
      return rtn
    end
    list.each do |ii|
      if(ii.type_of? Range)
        ii = ii.to_wrapper
        ii.re_range!(ng)
      end
      if ((ii.type_of? Range) || (ii.type_of? Array))
        ii.each do |idx|
          rtn.push idx
        end
      else 
        rtn.push ii
      end
    end
    return rtn
  end
  def self.copy(dest, source, indexes)
    rtn = []
    if indexes == :all
      if source.type_of? Array
        source.count.times do |idx|
          rtn.push(dest[idx]=source[idx])
        end
      else
        dest.count.times do |idx|
          rtn.push(dest[idx]=source)
        end
      end
    else
      if source.type_of? Array
        ii = 0
        indexes.each do |idx|
          rtn.push(dest[idx]=source[ii])
          ii += 1
        end      
      else
        indexes.each do |idx|
          rtn.push(dest[idx]=source)
        end            
      end
    end
    rtn.count <= 1 ? rtn.first : rtn   
  end  
end


### need to figure out how left and right things work ...
####
###   OK this is cool,  things on the right if more than one go in one array
###   ... or a non-array if just one item

class XArray < ValueAdd
  LAM_ISORT = lambda {|a,b| a.inspect <=> b.inspect }
  LAM_SSORT = lambda {|a,b| a.to_s <=> b.to_s }
  capture_base_methods(Array)
  
  def initialize(obj=[])
    obj = obj.prim_value
    ensure_valid(obj)
    @value = obj
  end
  
  def valid_type(prm)
    return true if prm.kind_of? Array
    return true if prm.kind_of? XArray
    false
  end
  def isort 
    @value.sort &LAM_ISORT
  end
  def isort!
    @value.sort! &LAM_ISORT
  end
  def ssort
    @value.sort &LAM_SSORT
  end
  def ssort!
    @value.sort! &LAM_SSORT
  end
  def [](*list)
    list = PrimitiveWrapper::get_list(list,size)
    if list.empty?
      return nil
    elsif list.count==1
      if list.first == :all
        return @value
      elsif list.first.type_of? Range  # get_list may pick this appart
        rtn = []
        list.first.to_wrapper.re_range(size).each do |idx|
          rtn.push @value[idx]
        end
        return rtn  
      else
        return @value[list.first]
      end
    end
    rtn = []
    # for now keep this here until you can justify that it is not needed
    list.each do |ii|   # I don't need to do this do I ???  get_list should untangle things
      if ii.type_of? Range
        ii.to_wrapper.re_range(size).each do |idx|
          rtn.push @value[idx]
        end
      elsif ii.type_of? Array
        ii.each do |idx|
          rtn.push @value[idx]
        end
      else
        rtn.push @value[ii]
      end
    end
    return rtn
  end
  def []=(*list)
    data = list.pop
    list = PrimitiveWrapper::get_list(list,size)
    return nil if list.empty?
    if list.count==1
      if list.first.respond_to? :each
        list = list.first
        if list.type_of? Range
          list = list.to_xr.re_range(size)
        end
      end
    end
    if list.count==1
      if list.first == :all
        PrimitiveWrapper.copy(@value, data, :all) 
      end
      idxs = list.first
      if (idxs.respond_to? :each)
        if idxs.type_of? Range
          idxs = idxs.to_xr.re_range(size)
        end        
        idxs.each do |idx|
          @value[idx] = data
        end
      else
        @value[idxs] = data
      end 
    else
      if data.type_of? Array
        data = PrimitiveWrapper::get_list(data,size)
      end
      PrimitiveWrapper.copy(@value, data, list)
    end
  end
  def delete_at(*index_list)
    return nil if index_list.empty?
    rtn = []
    if (index_list.count==1)
      if index_list.first==:all
        rtn = @value.dup
        @value.empty!
        return rtn
      else
        ii = index_list.first
        rtn = self[ii]
        if ii.type_of? Range
          ii = ii.to_xr
          ii.re_range!(size)
          ii.reorder!
          ii.reverse_each do |idx|
            @value.delete_at(idx)
          end
          return rtn
        end
      end
    end
    rtn = self[*index_list]
    list = PrimitiveWrapper::get_list(index_list,size)
    list = list & list  # remove duplicates
    list.sort.reverse.each do |idx|
      if idx.type_of? Range
        t_idx = idx.reorder
      else
        t_idx = idx
      end
      @value.delete_at(t_idx)
    end
    return rtn
  end
  def include?(*list)  # replace with list of stuff
    if list.count==1
      if list.first.respond_to? :each
        list = list.first
      end
    end
    list.each do |item|
      return false unless @value.include? item
    end
    true
  end
  def include_any?(*list)  # replace with list of stuff
    if list.count==1
      if list.first.respond_to? :each
        list = list.first
      end
    end
    list.each do |item|
      return true if @value.include? item
    end
    false
  end
end



